# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {}

modules = []

doc_url = "https://alexsax.github.ioomnidata-tools"

git_url = "https://github.com/alexsax/omnidata_tools/tree/main/"

def custom_doc_links(name): return None
